/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file RectBase.hpp
/// @desc Base class that defines and implements rectangle (box) rendering routines.
/// @edit 2024-08-02

#pragma once

#include <J3ML/LinearAlgebra.hpp>
#include <Event.h>
#include <EventConnection.h>
#include <Color4.hpp>

namespace JUI
{

    enum class BorderMode
    {
        Outline, /// As BorderWidth increases, the border grows outward. The dimensions of the widgets contents do not change.
        Middle,  /// As BorderWidth increases, the border grows evenly inward and outward. The dimensions of the widgets' contents are reduced at a 1:1 ratio.
        Inset    /// As BorderWidth increases, the border grows inward only. The dimensions of the widget's contents are reduced at a 1:2 ratio.
    };

    /// Base implementation for rendering rectangles with decorations.
    class RectBase {
    public:
        RectBase();
    public:
        Event<Vector2> MouseEnter;
        Event<Vector2> MouseExit;
        //Event<Vector2> MousePress;
        //Event<Vector2, bool> MouseRelease;
    public:
        void SetCornerRounding(float radius);
        void SetCornerRounding(float tlRadius, float trRadius, float blRadius, float brRadius);
        void SetCornerRoundingTL(float radius);
        void SetCornerRoundingTR(float radius);
        void SetCornerRoundingBL(float radius);
        void SetCornerRoundingBR(float radius);

        void SetClipsDescendants(bool clipping);
        void BGColor(const Color4& col);
        void BorderColor(const Color4& col);
        void SetBorderWidth(float w);


        Color4 BGColor() const;
        Color4 GetBorderColor() const;
        float GetBorderWidth() const;
        BorderMode GetBorderMode() const;

        void SetBorderMode(const BorderMode& mode)
        {
            border_mode = mode;
        }

        bool GetClipsDescendants() const;

        void SetBorderStyling(const Color4& color, float width);

        void Draw(const Vector2& pos, const Vector2& size);

    protected:
        BorderMode border_mode = BorderMode::Middle;
        bool mouse_press_debounce;
        bool mouse_inside_debounce;
        float border_width = 1.f;
        Color4 bg_color = {128,128,128, 255};
        Color4 border_color = {192, 192, 192, 0};
        bool clips_descendants = false; // Controls if child objects can render outside of their parent's rectangle bounds.
        float corner_rounding_radius = 0.f; // Curves the rectangle corners by N degrees.
    };
}